from __future__ import annotations

import json
from pathlib import Path
import sys
import unittest
from unittest.mock import patch

from _support import FakeResponse, FakeSession

ROOT = Path(__file__).resolve().parents[1]
sys.path.insert(0, str(ROOT / "src"))

from driftgate_sdk.client import DriftGateClient  # noqa: E402


def _run_response(state: str = "queued", run_id: str = "run_demo_1") -> dict:
    return {
        "run": {
            "id": run_id,
            "workspaceId": "ws_demo_1",
            "workflowVersionId": "wv_demo_1",
            "state": state,
            "requestedBy": "sdk-user-1",
            "requestedAt": "2026-01-01T00:00:00.000Z",
            "correlationId": "corr-123",
            "idempotencyKey": "idem-123",
            "triggerSource": "sdk",
        },
        "approval": None,
        "blocked": False,
        "policyDecisions": [],
        "entitlementDecision": {
            "id": "ent_1",
            "reasonCode": "within_limit",
            "reasonText": "allowed",
            "entitled": True,
        },
        "usageEntry": {"id": "usage_1", "quantity": 1},
    }


class DriftGateClientRunsTest(unittest.TestCase):
    def test_run_status_events_wait_for_terminal_and_approvals(self) -> None:
        session = FakeSession(
            [
                FakeResponse(status_code=201, payload=_run_response("queued", "run_001")),
                FakeResponse(status_code=200, payload=_run_response("queued", "run_001")),
                FakeResponse(
                    status_code=200,
                    payload={
                        "events": [
                            {
                                "id": "evt_1",
                                "runId": "run_001",
                                "type": "run.created",
                                "payload": {"state": "queued"},
                                "createdAt": "2026-01-01T00:00:01.000Z",
                            }
                        ]
                    },
                ),
                FakeResponse(status_code=200, payload=_run_response("queued", "run_001")),
                FakeResponse(status_code=200, payload=_run_response("succeeded", "run_001")),
                FakeResponse(
                    status_code=200,
                    payload={
                        "approvals": [
                            {
                                "approval": {
                                    "id": "ap_1",
                                    "runId": "run_001",
                                    "requiredRole": "owner",
                                    "status": "pending",
                                    "createdAt": "2026-01-01T00:00:00.000Z",
                                },
                                "run": _run_response("waiting_approval", "run_001")["run"],
                            }
                        ]
                    },
                ),
            ]
        )

        client = DriftGateClient(
            base_url="https://api.driftgate.ai",
            api_key="dgk_test_123",
            session=session,
        )

        created = client.run(
            workspace_id="ws_demo_1",
            workflow_version_id="wv_demo_1",
            input_payload={"amount": 100},
            requires_approval=True,
            required_role="owner",
            idempotency_key="idem-123",
            correlation_id="corr-123",
        )
        self.assertEqual(created.run.id, "run_001")
        self.assertEqual(created.run.trigger_source, "sdk")

        posted_payload = json.loads(session.calls[0]["data"])
        self.assertEqual(posted_payload["idempotencyKey"], "idem-123")
        self.assertEqual(posted_payload["correlationId"], "corr-123")
        self.assertEqual(posted_payload["triggerSource"], "sdk")
        self.assertEqual(session.calls[0]["headers"]["x-driftgate-api-key"], "dgk_test_123")

        status = client.status("run_001")
        self.assertEqual(status.run.state, "queued")

        events = client.events("run_001")
        self.assertEqual(len(events), 1)
        self.assertEqual(events[0].event_type, "run.created")

        with patch("driftgate_sdk.client.time.sleep", return_value=None):
            terminal = client.wait_for_terminal(
                "run_001",
                timeout_seconds=5,
                poll_interval_seconds=0.01,
            )
        self.assertEqual(terminal.run.state, "succeeded")

        approvals = client.approvals_list("ws_demo_1", status="pending")
        self.assertEqual(len(approvals), 1)
        self.assertEqual(approvals[0].approval.id, "ap_1")
        self.assertEqual(approvals[0].run.id, "run_001")


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import json
import time
from dataclasses import dataclass
from typing import Any, Dict, Optional

import requests

from .auth import AuthConfig
from .errors import DriftGateSDKError, build_error
from .models import (
    ApprovalQueueEntry,
    CanonicalResponse,
    EphemeralExecuteData,
    RunEvent,
    RunResponse,
    SessionResource,
    TriggerSource,
    parse_approvals,
    parse_run_events,
    parse_run_response,
    parse_v4_ephemeral_execute_envelope,
    parse_v4_execution_envelope,
    parse_v4_session_start_envelope,
)
from .retries import RetryPolicy


TERMINAL_STATES = {"succeeded", "failed", "denied", "timed_out", "canceled", "aborted"}


@dataclass
class DriftGateSessionHandle:
    _client: "DriftGateClient"
    session: SessionResource
    start_envelope: CanonicalResponse[Any]

    @property
    def session_id(self) -> str:
        return self.session.session_id

    @property
    def raw_envelope(self) -> Dict[str, Any]:
        return self.start_envelope.raw

    def execute(
        self,
        *,
        input: Dict[str, Any],
        policy: Optional[Dict[str, Any]] = None,
        route: Optional[Dict[str, Any]] = None,
        risk: Optional[Dict[str, Any]] = None,
        workflow_version_id: Optional[str] = None,
    ) -> CanonicalResponse[RunResponse]:
        payload: Dict[str, Any] = {"input": input}
        if policy is not None:
            payload["policy"] = policy
        if route is not None:
            payload["route"] = route
        if risk is not None:
            payload["risk"] = risk
        if workflow_version_id is not None:
            payload["workflowVersionId"] = workflow_version_id
        return self._client.execute_session(self.session_id, payload)


class _SessionNamespace:
    def __init__(self, client: "DriftGateClient") -> None:
        self._client = client

    def start(
        self,
        *,
        agent: str,
        workspace_id: Optional[str] = None,
        subject: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
        policy: Optional[Dict[str, Any]] = None,
        route: Optional[Dict[str, Any]] = None,
        risk: Optional[Dict[str, Any]] = None,
        workflow_version_id: Optional[str] = None,
        expires_at: Optional[str] = None,
    ) -> DriftGateSessionHandle:
        payload: Dict[str, Any] = {"agent": agent}
        if workspace_id is not None:
            payload["workspaceId"] = workspace_id
        if subject is not None:
            payload["subject"] = subject
        if metadata is not None:
            payload["metadata"] = metadata
        if policy is not None:
            payload["policy"] = policy
        if route is not None:
            payload["route"] = route
        if risk is not None:
            payload["risk"] = risk
        if workflow_version_id is not None:
            payload["workflowVersionId"] = workflow_version_id
        if expires_at is not None:
            payload["expiresAt"] = expires_at

        envelope = self._client.start_session(payload)
        if not envelope.ok or envelope.data is None:
            raise DriftGateSDKError(
                code=envelope.error.code if envelope.error else "INTERNAL",
                message=envelope.error.message if envelope.error else "session start failed",
                status_code=envelope.error.status if envelope.error else 500,
                correlation_id=envelope.meta.request_id,
                details=envelope.error.details if envelope.error else None,
            )

        return DriftGateSessionHandle(
            _client=self._client,
            session=envelope.data.session,
            start_envelope=envelope,
        )


class DriftGateClient:
    def __init__(
        self,
        *,
        base_url: str,
        api_key: Optional[str] = None,
        bearer_token: Optional[str] = None,
        timeout_seconds: float = 30.0,
        retry_policy: Optional[RetryPolicy] = None,
        session: Optional[requests.Session] = None,
    ) -> None:
        self._base_url = base_url.rstrip("/")
        self._auth = AuthConfig(api_key=api_key, bearer_token=bearer_token)
        self._timeout_seconds = timeout_seconds
        self._retry_policy = retry_policy or RetryPolicy()
        self._session = session or requests.Session()
        self.session = _SessionNamespace(self)

    def start_session(self, payload: Dict[str, Any]) -> CanonicalResponse[Any]:
        response = self._request("POST", "/v4/sessions.start", payload=payload, allow_retry=False)
        return parse_v4_session_start_envelope(response)

    def execute_session(self, session_id: str, payload: Dict[str, Any]) -> CanonicalResponse[RunResponse]:
        response = self._request(
            "POST",
            f"/v4/sessions/{session_id}/executions.execute",
            payload=payload,
            allow_retry=False,
        )
        return parse_v4_execution_envelope(response)

    def execute(
        self,
        *,
        agent: str,
        input: Dict[str, Any],
        workspace_id: Optional[str] = None,
        subject: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
        policy: Optional[Dict[str, Any]] = None,
        route: Optional[Dict[str, Any]] = None,
        risk: Optional[Dict[str, Any]] = None,
        workflow_version_id: Optional[str] = None,
    ) -> CanonicalResponse[EphemeralExecuteData]:
        payload: Dict[str, Any] = {
            "agent": agent,
            "input": input,
        }
        if workspace_id is not None:
            payload["workspaceId"] = workspace_id
        if subject is not None:
            payload["subject"] = subject
        if metadata is not None:
            payload["metadata"] = metadata
        if policy is not None:
            payload["policy"] = policy
        if route is not None:
            payload["route"] = route
        if risk is not None:
            payload["risk"] = risk
        if workflow_version_id is not None:
            payload["workflowVersionId"] = workflow_version_id

        response = self._request("POST", "/v4/execute", payload=payload, allow_retry=False)
        return parse_v4_ephemeral_execute_envelope(response)

    def run(
        self,
        *,
        workspace_id: str,
        workflow_version_id: str,
        input_payload: Optional[Dict[str, Any]] = None,
        requires_approval: Optional[bool] = None,
        required_role: Optional[str] = None,
        sla_policy_id: Optional[str] = None,
        idempotency_key: Optional[str] = None,
        correlation_id: Optional[str] = None,
        trigger_source: TriggerSource = "sdk",
    ) -> RunResponse:
        payload: Dict[str, Any] = {
            "workspaceId": workspace_id,
            "workflowVersionId": workflow_version_id,
            "triggerSource": trigger_source,
        }
        if input_payload is not None:
            payload["input"] = input_payload
        if requires_approval is not None:
            payload["requiresApproval"] = requires_approval
        if required_role is not None:
            payload["requiredRole"] = required_role
        if sla_policy_id is not None:
            payload["slaPolicyId"] = sla_policy_id
        if idempotency_key is not None:
            payload["idempotencyKey"] = idempotency_key
        if correlation_id is not None:
            payload["correlationId"] = correlation_id

        response = self._request("POST", "/v1/headless/runs", payload=payload, allow_retry=False)
        return parse_run_response(response)

    def status(self, run_id: str) -> RunResponse:
        response = self._request("GET", f"/v1/headless/runs/{run_id}")
        return parse_run_response(response)

    def events(self, run_id: str) -> list[RunEvent]:
        response = self._request("GET", f"/v1/headless/runs/{run_id}/events")
        return parse_run_events(response)

    def wait_for_terminal(
        self,
        run_id: str,
        *,
        timeout_seconds: float = 120.0,
        poll_interval_seconds: float = 1.5,
    ) -> RunResponse:
        started = time.monotonic()
        while True:
            current = self.status(run_id)
            if current.run.state in TERMINAL_STATES:
                return current
            if (time.monotonic() - started) >= timeout_seconds:
                raise DriftGateSDKError(
                    code="timeout",
                    message=f"run {run_id} did not reach terminal state before timeout",
                    status_code=408,
                )
            time.sleep(poll_interval_seconds)

    def approvals_list(
        self,
        workspace_id: str,
        *,
        status: Optional[str] = None,
    ) -> list[ApprovalQueueEntry]:
        query = f"?status={status}" if status else ""
        response = self._request(
            "GET",
            f"/v1/headless/workspaces/{workspace_id}/approvals{query}",
        )
        return parse_approvals(response)

    def approve(self, approval_id: str) -> RunResponse:
        response = self._request(
            "POST",
            f"/v1/headless/approvals/{approval_id}/approve",
            allow_retry=False,
        )
        return parse_run_response(response)

    def deny(self, approval_id: str) -> RunResponse:
        response = self._request(
            "POST",
            f"/v1/headless/approvals/{approval_id}/deny",
            allow_retry=False,
        )
        return parse_run_response(response)

    def _request(
        self,
        method: str,
        path: str,
        *,
        payload: Optional[Dict[str, Any]] = None,
        allow_retry: bool = True,
    ) -> Dict[str, Any]:
        url = f"{self._base_url}{path}"
        headers = {"content-type": "application/json", **self._auth.build_headers()}

        attempt = 1
        while True:
            try:
                response = self._session.request(
                    method=method,
                    url=url,
                    headers=headers,
                    data=json.dumps(payload) if payload is not None else None,
                    timeout=self._timeout_seconds,
                )
            except requests.RequestException as error:
                if not allow_retry or attempt >= self._retry_policy.max_attempts:
                    raise DriftGateSDKError(
                        code="network_error",
                        message=str(error),
                        status_code=0,
                    ) from error
                self._retry_policy.sleep(attempt)
                attempt += 1
                continue

            if response.ok:
                if response.text.strip():
                    parsed = response.json()
                    if isinstance(parsed, dict):
                        return parsed
                    return {"value": parsed}
                return {}

            envelope = _parse_error_envelope(response)
            if (
                allow_retry
                and self._retry_policy.should_retry(attempt, response.status_code)
                and method.upper() in {"GET", "HEAD"}
            ):
                self._retry_policy.sleep(attempt)
                attempt += 1
                continue
            raise envelope


def _parse_error_envelope(response: requests.Response) -> DriftGateSDKError:
    try:
        payload = response.json()
    except ValueError:
        payload = None

    if (
        isinstance(payload, dict)
        and payload.get("ok") is False
        and isinstance(payload.get("error"), dict)
        and isinstance(payload.get("meta"), dict)
    ):
        error_payload = payload["error"]
        meta_payload = payload["meta"]
        return build_error(
            status_code=int(error_payload.get("status", response.status_code)),
            code=str(error_payload.get("code", "INTERNAL")),
            message=str(error_payload.get("message", "request failed")),
            correlation_id=str(meta_payload.get("requestId"))
            if isinstance(meta_payload.get("requestId"), str)
            else None,
            details=error_payload.get("details"),
        )

    if isinstance(payload, dict) and "code" in payload and "message" in payload:
        return build_error(
            status_code=response.status_code,
            code=str(payload.get("code")),
            message=str(payload.get("message")),
            correlation_id=str(payload["correlation_id"])
            if isinstance(payload.get("correlation_id"), str)
            else None,
            details=payload.get("details"),
        )

    if isinstance(payload, dict) and "error" in payload and "message" in payload:
        return build_error(
            status_code=response.status_code,
            code=str(payload.get("error")),
            message=str(payload.get("message")),
            details=payload.get("issues"),
        )

    return DriftGateSDKError(
        code="http_error",
        message=f"request failed ({response.status_code}): {response.text}",
        status_code=response.status_code,
    )



from __future__ import annotations

from pathlib import Path
import sys
import unittest
from unittest.mock import patch

from _support import FakeResponse, FakeSession

ROOT = Path(__file__).resolve().parents[1]
sys.path.insert(0, str(ROOT / "src"))

from driftgate_sdk.client import DriftGateClient  # noqa: E402
from driftgate_sdk.errors import (  # noqa: E402
    ForbiddenError,
    RateLimitedError,
    UnauthorizedError,
)
from driftgate_sdk.retries import RetryPolicy  # noqa: E402


def _status_payload(state: str = "queued") -> dict:
    return {
        "run": {
            "id": "run_123",
            "workspaceId": "ws_demo_1",
            "workflowVersionId": "wv_demo_1",
            "state": state,
            "requestedBy": "sdk-user-1",
            "requestedAt": "2026-01-01T00:00:00.000Z",
            "correlationId": "corr-123",
            "idempotencyKey": None,
            "triggerSource": "sdk",
        },
        "approval": None,
    }


class DriftGateClientErrorTest(unittest.TestCase):
    def test_unauthorized_error_mapping(self) -> None:
        session = FakeSession(
            [
                FakeResponse(
                    status_code=401,
                    payload={
                        "code": "unauthorized",
                        "message": "missing session token",
                        "correlation_id": "corr-auth",
                    },
                )
            ]
        )
        client = DriftGateClient(
            base_url="https://api.driftgate.ai",
            bearer_token="token_123",
            session=session,
        )

        with self.assertRaises(UnauthorizedError) as context:
            client.status("run_123")
        self.assertEqual(context.exception.status_code, 401)
        self.assertEqual(context.exception.correlation_id, "corr-auth")

    def test_legacy_error_envelope_mapping(self) -> None:
        session = FakeSession(
            [
                FakeResponse(
                    status_code=403,
                    payload={
                        "error": "forbidden",
                        "message": "workspace scope denied",
                        "issues": {"workspaceId": "ws_blocked"},
                    },
                )
            ]
        )
        client = DriftGateClient(
            base_url="https://api.driftgate.ai",
            bearer_token="token_123",
            session=session,
        )

        with self.assertRaises(ForbiddenError) as context:
            client.status("run_123")
        self.assertEqual(context.exception.code, "forbidden")
        self.assertEqual(context.exception.details["workspaceId"], "ws_blocked")

    def test_rate_limit_retries_for_get_then_succeeds(self) -> None:
        session = FakeSession(
            [
                FakeResponse(
                    status_code=429,
                    payload={
                        "code": "rate_limit_exceeded",
                        "message": "api key rate limit exceeded",
                    },
                ),
                FakeResponse(status_code=200, payload=_status_payload("queued")),
            ]
        )
        client = DriftGateClient(
            base_url="https://api.driftgate.ai",
            api_key="dgk_test_123",
            session=session,
            retry_policy=RetryPolicy(max_attempts=2, base_backoff_seconds=0.0, jitter_seconds=0.0),
        )

        with patch("driftgate_sdk.retries.time.sleep", return_value=None):
            status = client.status("run_123")
        self.assertEqual(status.run.id, "run_123")
        self.assertEqual(len(session.calls), 2)

    def test_rate_limit_without_retry_path_raises(self) -> None:
        session = FakeSession(
            [
                FakeResponse(
                    status_code=429,
                    payload={
                        "code": "rate_limit_exceeded",
                        "message": "api key rate limit exceeded",
                    },
                )
            ]
        )
        client = DriftGateClient(
            base_url="https://api.driftgate.ai",
            api_key="dgk_test_123",
            session=session,
        )

        with self.assertRaises(RateLimitedError):
            client.run(workspace_id="ws_demo_1", workflow_version_id="wv_demo_1")


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

from dataclasses import dataclass
import random
import time


@dataclass(frozen=True)
class RetryPolicy:
    max_attempts: int = 3
    base_backoff_seconds: float = 0.25
    max_backoff_seconds: float = 2.0
    jitter_seconds: float = 0.1

    def should_retry(self, attempt: int, status_code: int) -> bool:
        if attempt >= self.max_attempts:
            return False
        return status_code == 429 or status_code >= 500

    def sleep(self, attempt: int) -> None:
        exponential = self.base_backoff_seconds * (2 ** max(attempt - 1, 0))
        bounded = min(exponential, self.max_backoff_seconds)
        jitter = random.uniform(0.0, self.jitter_seconds)
        time.sleep(bounded + jitter)



from .client import DriftGateClient, DriftGateSessionHandle
from .errors import (
  ConflictError,
  DriftGateSDKError,
  ForbiddenError,
    InvalidRequestError,
    NotFoundError,
  RateLimitedError,
  UnauthorizedError,
)
from .models import CanonicalResponse

__all__ = [
  "ConflictError",
  "CanonicalResponse",
  "DriftGateClient",
  "DriftGateSessionHandle",
  "DriftGateSDKError",
  "ForbiddenError",
  "InvalidRequestError",
    "NotFoundError",
    "RateLimitedError",
    "UnauthorizedError",
]



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Optional


@dataclass
class DriftGateSDKError(Exception):
    code: str
    message: str
    status_code: int
    correlation_id: Optional[str] = None
    details: Optional[Any] = None

    def __str__(self) -> str:
        return f"{self.code} ({self.status_code}): {self.message}"


class InvalidRequestError(DriftGateSDKError):
    pass


class UnauthorizedError(DriftGateSDKError):
    pass


class ForbiddenError(DriftGateSDKError):
    pass


class NotFoundError(DriftGateSDKError):
    pass


class ConflictError(DriftGateSDKError):
    pass


class RateLimitedError(DriftGateSDKError):
    pass


ERROR_CODE_MAP = {
    "invalid_request": InvalidRequestError,
    "unauthorized": UnauthorizedError,
    "forbidden": ForbiddenError,
    "not_found": NotFoundError,
    "conflict": ConflictError,
    "rate_limit_exceeded": RateLimitedError,
    "AUTH_INVALID": UnauthorizedError,
    "POLICY_DENIED": ForbiddenError,
    "RISK_EXCEEDED": ForbiddenError,
    "ROUTE_UNAVAILABLE": NotFoundError,
    "TOOL_BLOCKED": ForbiddenError,
    "RATE_LIMITED": RateLimitedError,
    "TIMEOUT": DriftGateSDKError,
    "INTERNAL": DriftGateSDKError,
}


def build_error(
    *,
    status_code: int,
    code: str,
    message: str,
    correlation_id: Optional[str] = None,
    details: Optional[Any] = None,
) -> DriftGateSDKError:
    error_type = ERROR_CODE_MAP.get(code, DriftGateSDKError)
    return error_type(
        code=code,
        message=message,
        status_code=status_code,
        correlation_id=correlation_id,
        details=details,
    )



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, Generic, List, Literal, Optional, TypeVar


TriggerSource = Literal["ui", "api", "sdk", "cli", "hosted", "webhook"]
RunState = Literal[
    "queued",
    "running",
    "waiting_approval",
    "approved",
    "denied",
    "succeeded",
    "failed",
    "aborted",
    "timed_out",
    "canceled",
]

CanonicalErrorCode = Literal[
    "AUTH_INVALID",
    "POLICY_DENIED",
    "RISK_EXCEEDED",
    "ROUTE_UNAVAILABLE",
    "TOOL_BLOCKED",
    "RATE_LIMITED",
    "TIMEOUT",
    "INTERNAL",
]


@dataclass(frozen=True)
class CanonicalPolicyRef:
    ref: str
    version: str


@dataclass(frozen=True)
class CanonicalRouteRef:
    provider: Optional[str] = None
    model: Optional[str] = None
    region: Optional[str] = None


@dataclass(frozen=True)
class CanonicalRiskMeta:
    score: Optional[float] = None
    decision: Optional[Literal["allow", "deny", "review"]] = None


@dataclass(frozen=True)
class CanonicalTimingMs:
    total: float
    policy: Optional[float] = None
    route: Optional[float] = None
    tool: Optional[float] = None


@dataclass(frozen=True)
class CanonicalMeta:
    request_id: str
    timing_ms: CanonicalTimingMs
    session_id: Optional[str] = None
    execution_id: Optional[str] = None
    lineage_id: Optional[str] = None
    policy: Optional[CanonicalPolicyRef] = None
    route: Optional[CanonicalRouteRef] = None
    risk: Optional[CanonicalRiskMeta] = None

    @property
    def requestId(self) -> str:
        return self.request_id

    @property
    def timingMs(self) -> CanonicalTimingMs:
        return self.timing_ms

    @property
    def sessionId(self) -> Optional[str]:
        return self.session_id

    @property
    def executionId(self) -> Optional[str]:
        return self.execution_id

    @property
    def lineageId(self) -> Optional[str]:
        return self.lineage_id


@dataclass(frozen=True)
class CanonicalError:
    code: str
    message: str
    status: int
    retryable: bool
    details: Optional[Dict[str, Any]] = None


T = TypeVar("T")


@dataclass(frozen=True)
class CanonicalResponse(Generic[T]):
    ok: bool
    data: Optional[T]
    meta: CanonicalMeta
    error: Optional[CanonicalError]
    raw: Dict[str, Any]


@dataclass(frozen=True)
class SessionResource:
    session_id: str
    workspace_id: str
    agent: str
    created_at: str
    subject: Optional[str] = None
    metadata: Optional[Dict[str, Any]] = None
    policy: Optional[CanonicalPolicyRef] = None
    route: Optional[CanonicalRouteRef] = None
    risk: Optional[CanonicalRiskMeta] = None
    workflow_version_id: Optional[str] = None
    expires_at: Optional[str] = None

    @property
    def sessionId(self) -> str:
        return self.session_id


@dataclass(frozen=True)
class SessionStartData:
    session: SessionResource


@dataclass(frozen=True)
class EphemeralExecuteData:
    session: SessionResource
    execution: "RunResponse"


@dataclass(frozen=True)
class RunSummary:
    id: str
    workspace_id: str
    workflow_version_id: str
    state: RunState
    requested_by: str
    requested_at: str
    correlation_id: str
    trigger_source: TriggerSource
    idempotency_key: Optional[str] = None
    updated_at: Optional[str] = None
    started_at: Optional[str] = None
    completed_at: Optional[str] = None
    sla_policy_id: Optional[str] = None
    sla_due_at: Optional[str] = None
    sla_violated_at: Optional[str] = None


@dataclass(frozen=True)
class ApprovalSummary:
    id: str
    run_id: str
    status: Literal["pending", "approved", "denied"]
    requested_at: Optional[str] = None
    required_role: Optional[str] = None
    decided_at: Optional[str] = None
    decided_by: Optional[str] = None


@dataclass(frozen=True)
class EntitlementDecision:
    id: str
    reason_code: str
    reason_text: str
    entitled: bool


@dataclass(frozen=True)
class UsageEntry:
    id: str
    quantity: int


@dataclass(frozen=True)
class RunEvent:
    id: str
    run_id: str
    event_type: str
    payload: Dict[str, Any]
    created_at: str


@dataclass(frozen=True)
class ApprovalQueueEntry:
    approval: ApprovalSummary
    run: Optional[RunSummary] = None


@dataclass(frozen=True)
class RunResponse:
    run: RunSummary
    approval: Optional[ApprovalSummary] = None
    blocked: Optional[bool] = None
    policy_decisions: Optional[List[Dict[str, Any]]] = None
    entitlement_decision: Optional[EntitlementDecision] = None
    usage_entry: Optional[UsageEntry] = None


def parse_run_summary(payload: Dict[str, Any]) -> RunSummary:
    return RunSummary(
        id=str(payload["id"]),
        workspace_id=str(payload["workspaceId"]),
        workflow_version_id=str(payload["workflowVersionId"]),
        state=payload["state"],
        requested_by=str(payload["requestedBy"]),
        requested_at=str(payload["requestedAt"]),
        correlation_id=str(payload["correlationId"]),
        trigger_source=payload["triggerSource"],
        idempotency_key=payload.get("idempotencyKey"),
        updated_at=payload.get("updatedAt"),
        started_at=payload.get("startedAt"),
        completed_at=payload.get("completedAt"),
        sla_policy_id=payload.get("slaPolicyId"),
        sla_due_at=payload.get("slaDueAt"),
        sla_violated_at=payload.get("slaViolatedAt"),
    )


def parse_approval_summary(payload: Dict[str, Any]) -> ApprovalSummary:
    return ApprovalSummary(
        id=str(payload["id"]),
        run_id=str(payload["runId"]),
        status=payload["status"],
        requested_at=payload.get("requestedAt") or payload.get("createdAt"),
        required_role=payload.get("requiredRole"),
        decided_at=payload.get("decidedAt"),
        decided_by=payload.get("decidedBy"),
    )


def parse_run_response(payload: Dict[str, Any]) -> RunResponse:
    entitlement = payload.get("entitlementDecision")
    usage = payload.get("usageEntry")
    approval = payload.get("approval")
    return RunResponse(
        run=parse_run_summary(payload["run"]),
        approval=parse_approval_summary(approval) if isinstance(approval, dict) and approval else None,
        blocked=payload.get("blocked"),
        policy_decisions=payload.get("policyDecisions"),
        entitlement_decision=EntitlementDecision(
            id=str(entitlement["id"]),
            reason_code=str(entitlement["reasonCode"]),
            reason_text=str(entitlement["reasonText"]),
            entitled=bool(entitlement["entitled"]),
        )
        if isinstance(entitlement, dict)
        else None,
        usage_entry=UsageEntry(id=str(usage["id"]), quantity=int(usage["quantity"]))
        if isinstance(usage, dict)
        else None,
    )


def parse_run_events(payload: Dict[str, Any]) -> List[RunEvent]:
    events: List[RunEvent] = []
    for raw in payload.get("events", []):
        if not isinstance(raw, dict):
            continue
        events.append(
            RunEvent(
                id=str(raw.get("id", "")),
                run_id=str(raw.get("runId", "")),
                event_type=str(raw.get("type", "")),
                payload=raw.get("payload", {}) if isinstance(raw.get("payload"), dict) else {},
                created_at=str(raw.get("createdAt", "")),
            )
        )
    return events


def parse_approvals(payload: Dict[str, Any]) -> List[ApprovalQueueEntry]:
    rows = payload.get("approvals", [])
    parsed: List[ApprovalQueueEntry] = []
    for row in rows:
        if not isinstance(row, dict):
            continue
        if isinstance(row.get("approval"), dict):
            approval = parse_approval_summary(row["approval"])
            run = parse_run_summary(row["run"]) if isinstance(row.get("run"), dict) else None
            parsed.append(ApprovalQueueEntry(approval=approval, run=run))
            continue
        parsed.append(ApprovalQueueEntry(approval=parse_approval_summary(row), run=None))
    return parsed


def parse_canonical_meta(payload: Dict[str, Any]) -> CanonicalMeta:
    timing = payload.get("timingMs") if isinstance(payload.get("timingMs"), dict) else {}
    policy_payload = payload.get("policy") if isinstance(payload.get("policy"), dict) else None
    route_payload = payload.get("route") if isinstance(payload.get("route"), dict) else None
    risk_payload = payload.get("risk") if isinstance(payload.get("risk"), dict) else None

    policy = (
        CanonicalPolicyRef(
            ref=str(policy_payload.get("ref", "")),
            version=str(policy_payload.get("version", "")),
        )
        if policy_payload
        else None
    )
    route = (
        CanonicalRouteRef(
            provider=str(route_payload["provider"]) if route_payload.get("provider") is not None else None,
            model=str(route_payload["model"]) if route_payload.get("model") is not None else None,
            region=str(route_payload["region"]) if route_payload.get("region") is not None else None,
        )
        if route_payload
        else None
    )
    risk = (
        CanonicalRiskMeta(
            score=float(risk_payload["score"]) if risk_payload.get("score") is not None else None,
            decision=str(risk_payload["decision"]) if risk_payload.get("decision") is not None else None,
        )
        if risk_payload
        else None
    )

    return CanonicalMeta(
        request_id=str(payload.get("requestId", "")),
        session_id=str(payload["sessionId"]) if payload.get("sessionId") is not None else None,
        execution_id=str(payload["executionId"]) if payload.get("executionId") is not None else None,
        lineage_id=str(payload["lineageId"]) if payload.get("lineageId") is not None else None,
        policy=policy,
        route=route,
        risk=risk,
        timing_ms=CanonicalTimingMs(
            total=float(timing.get("total", 0)),
            policy=float(timing["policy"]) if timing.get("policy") is not None else None,
            route=float(timing["route"]) if timing.get("route") is not None else None,
            tool=float(timing["tool"]) if timing.get("tool") is not None else None,
        ),
    )


def parse_session_resource(payload: Dict[str, Any]) -> SessionResource:
    policy_payload = payload.get("policy") if isinstance(payload.get("policy"), dict) else None
    route_payload = payload.get("route") if isinstance(payload.get("route"), dict) else None
    risk_payload = payload.get("risk") if isinstance(payload.get("risk"), dict) else None

    return SessionResource(
        session_id=str(payload["sessionId"]),
        workspace_id=str(payload["workspaceId"]),
        agent=str(payload["agent"]),
        created_at=str(payload["createdAt"]),
        subject=str(payload["subject"]) if payload.get("subject") is not None else None,
        metadata=payload.get("metadata") if isinstance(payload.get("metadata"), dict) else None,
        policy=CanonicalPolicyRef(ref=str(policy_payload["ref"]), version=str(policy_payload["version"]))
        if policy_payload
        else None,
        route=CanonicalRouteRef(
            provider=str(route_payload["provider"]) if route_payload.get("provider") is not None else None,
            model=str(route_payload["model"]) if route_payload.get("model") is not None else None,
            region=str(route_payload["region"]) if route_payload.get("region") is not None else None,
        )
        if route_payload
        else None,
        risk=CanonicalRiskMeta(
            score=float(risk_payload["score"]) if risk_payload.get("score") is not None else None,
            decision=str(risk_payload["decision"]) if risk_payload.get("decision") is not None else None,
        )
        if risk_payload
        else None,
        workflow_version_id=str(payload["workflowVersionId"]) if payload.get("workflowVersionId") is not None else None,
        expires_at=str(payload["expiresAt"]) if payload.get("expiresAt") is not None else None,
    )


def parse_canonical_error(payload: Dict[str, Any]) -> CanonicalError:
    details = payload.get("details") if isinstance(payload.get("details"), dict) else None
    return CanonicalError(
        code=str(payload.get("code", "INTERNAL")),
        message=str(payload.get("message", "unknown error")),
        status=int(payload.get("status", 500)),
        retryable=bool(payload.get("retryable", False)),
        details=details,
    )


def parse_v4_session_start_envelope(payload: Dict[str, Any]) -> CanonicalResponse[SessionStartData]:
    data_payload = payload.get("data") if isinstance(payload.get("data"), dict) else None
    session_payload = data_payload.get("session") if isinstance(data_payload, dict) else None
    parsed_data = SessionStartData(session=parse_session_resource(session_payload)) if isinstance(session_payload, dict) else None

    error_payload = payload.get("error") if isinstance(payload.get("error"), dict) else None
    parsed_error = parse_canonical_error(error_payload) if error_payload else None

    return CanonicalResponse(
        ok=bool(payload.get("ok", False)),
        data=parsed_data,
        meta=parse_canonical_meta(payload.get("meta", {}) if isinstance(payload.get("meta"), dict) else {}),
        error=parsed_error,
        raw=payload,
    )


def parse_v4_execution_envelope(payload: Dict[str, Any]) -> CanonicalResponse[RunResponse]:
    data_payload = payload.get("data") if isinstance(payload.get("data"), dict) else None
    parsed_data = parse_run_response(data_payload) if isinstance(data_payload, dict) and isinstance(data_payload.get("run"), dict) else None

    error_payload = payload.get("error") if isinstance(payload.get("error"), dict) else None
    parsed_error = parse_canonical_error(error_payload) if error_payload else None

    return CanonicalResponse(
        ok=bool(payload.get("ok", False)),
        data=parsed_data,
        meta=parse_canonical_meta(payload.get("meta", {}) if isinstance(payload.get("meta"), dict) else {}),
        error=parsed_error,
        raw=payload,
    )


def parse_v4_ephemeral_execute_envelope(payload: Dict[str, Any]) -> CanonicalResponse[EphemeralExecuteData]:
    data_payload = payload.get("data") if isinstance(payload.get("data"), dict) else None
    session_payload = data_payload.get("session") if isinstance(data_payload, dict) else None
    execution_payload = data_payload.get("execution") if isinstance(data_payload, dict) else None

    parsed_data: Optional[EphemeralExecuteData] = None
    if isinstance(session_payload, dict) and isinstance(execution_payload, dict):
        parsed_data = EphemeralExecuteData(
            session=parse_session_resource(session_payload),
            execution=parse_run_response(execution_payload),
        )

    error_payload = payload.get("error") if isinstance(payload.get("error"), dict) else None
    parsed_error = parse_canonical_error(error_payload) if error_payload else None

    return CanonicalResponse(
        ok=bool(payload.get("ok", False)),
        data=parsed_data,
        meta=parse_canonical_meta(payload.get("meta", {}) if isinstance(payload.get("meta"), dict) else {}),
        error=parsed_error,
        raw=payload,
    )



from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, Optional


@dataclass(frozen=True)
class AuthConfig:
    api_key: Optional[str] = None
    bearer_token: Optional[str] = None

    def build_headers(self) -> Dict[str, str]:
        if self.api_key:
            return {"x-driftgate-api-key": self.api_key}
        if self.bearer_token:
            return {"Authorization": f"Bearer {self.bearer_token}"}
        raise ValueError("either api_key or bearer_token is required")



from __future__ import annotations

from pathlib import Path
import sys
import unittest

from _support import FakeResponse, FakeSession

ROOT = Path(__file__).resolve().parents[1]
sys.path.insert(0, str(ROOT / "src"))

from driftgate_sdk.client import DriftGateClient  # noqa: E402
from driftgate_sdk.errors import InvalidRequestError  # noqa: E402


def _response_for(run_id: str) -> dict:
    return {
        "run": {
            "id": run_id,
            "workspaceId": "ws_demo_1",
            "workflowVersionId": "wv_demo_1",
            "state": "queued",
            "requestedBy": "sdk-user-1",
            "requestedAt": "2026-01-01T00:00:00.000Z",
            "correlationId": "corr-123",
            "idempotencyKey": "idem-xyz",
            "triggerSource": "sdk",
        },
        "approval": None,
        "blocked": False,
        "policyDecisions": [],
        "entitlementDecision": {
            "id": "ent_1",
            "reasonCode": "within_limit",
            "reasonText": "allowed",
            "entitled": True,
        },
        "usageEntry": {"id": "usage_1", "quantity": 1},
    }


class DriftGateClientIdempotencyTest(unittest.TestCase):
    def test_same_idempotency_key_same_payload_returns_same_run(self) -> None:
        session = FakeSession(
            [
                FakeResponse(status_code=201, payload=_response_for("run_same")),
                FakeResponse(status_code=201, payload=_response_for("run_same")),
            ]
        )
        client = DriftGateClient(
            base_url="https://api.driftgate.ai",
            api_key="dgk_test_123",
            session=session,
        )

        first = client.run(
            workspace_id="ws_demo_1",
            workflow_version_id="wv_demo_1",
            input_payload={"amount": 10},
            idempotency_key="idem-xyz",
            correlation_id="corr-a",
        )
        second = client.run(
            workspace_id="ws_demo_1",
            workflow_version_id="wv_demo_1",
            input_payload={"amount": 10},
            idempotency_key="idem-xyz",
            correlation_id="corr-b",
        )
        self.assertEqual(first.run.id, second.run.id)

    def test_same_idempotency_key_different_payload_raises_conflict_like_error(self) -> None:
        session = FakeSession(
            [
                FakeResponse(status_code=201, payload=_response_for("run_same")),
                FakeResponse(
                    status_code=400,
                    payload={
                        "code": "invalid_request",
                        "message": "idempotency key already used with a different request payload",
                        "correlation_id": "corr-conflict",
                    },
                ),
            ]
        )
        client = DriftGateClient(
            base_url="https://api.driftgate.ai",
            api_key="dgk_test_123",
            session=session,
        )

        client.run(
            workspace_id="ws_demo_1",
            workflow_version_id="wv_demo_1",
            input_payload={"amount": 10},
            idempotency_key="idem-xyz",
            correlation_id="corr-a",
        )
        with self.assertRaises(InvalidRequestError) as context:
            client.run(
                workspace_id="ws_demo_1",
                workflow_version_id="wv_demo_2",
                input_payload={"amount": 999},
                idempotency_key="idem-xyz",
                correlation_id="corr-b",
            )
        self.assertEqual(context.exception.code, "invalid_request")
        self.assertEqual(context.exception.correlation_id, "corr-conflict")


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import json
from pathlib import Path
import sys
import unittest

from _support import FakeResponse, FakeSession

ROOT = Path(__file__).resolve().parents[1]
sys.path.insert(0, str(ROOT / "src"))

from driftgate_sdk import DriftGateClient, DriftGateSDKError  # noqa: E402


class V4EnvelopeTests(unittest.TestCase):
    def test_session_start_and_execute_uses_camel_case_wire_keys(self) -> None:
        fake = FakeSession(
            [
                FakeResponse(
                    201,
                    {
                        "ok": True,
                        "data": {
                            "session": {
                                "sessionId": "sess_123",
                                "workspaceId": "ws_123",
                                "agent": "refund-agent",
                                "createdAt": "2026-02-28T00:00:00.000Z",
                            }
                        },
                        "meta": {
                            "requestId": "req_start",
                            "sessionId": "sess_123",
                            "timingMs": {"total": 12.2},
                        },
                        "error": None,
                    },
                ),
                FakeResponse(
                    201,
                    {
                        "ok": True,
                        "data": {
                            "run": {
                                "id": "run_123",
                                "workspaceId": "ws_123",
                                "workflowVersionId": "wv_123",
                                "state": "queued",
                                "requestedBy": "user_1",
                                "requestedAt": "2026-02-28T00:00:01.000Z",
                                "correlationId": "corr_123",
                                "triggerSource": "sdk",
                            },
                            "approval": None,
                            "blocked": False,
                            "policyDecisions": [],
                            "entitlementDecision": {
                                "id": "ent_1",
                                "reasonCode": "entitled",
                                "reasonText": "entitled",
                                "entitled": True,
                            },
                            "usageEntry": {"id": "usage_1", "quantity": 1},
                        },
                        "meta": {
                            "requestId": "req_exec",
                            "sessionId": "sess_123",
                            "executionId": "run_123",
                            "lineageId": "run_123",
                            "timingMs": {"total": 16.4},
                        },
                        "error": None,
                    },
                ),
            ]
        )
        client = DriftGateClient(base_url="https://api.driftgate.ai", bearer_token="token", session=fake)

        session = client.session.start(agent="refund-agent", workspace_id="ws_123")
        envelope = session.execute(input={"orderId": "123"})

        self.assertEqual(session.session_id, "sess_123")
        self.assertEqual(envelope.meta.execution_id, "run_123")
        self.assertEqual(len(fake.calls), 2)

        start_payload = json.loads(fake.calls[0]["data"])
        execute_payload = json.loads(fake.calls[1]["data"])
        self.assertEqual(start_payload, {"agent": "refund-agent", "workspaceId": "ws_123"})
        self.assertEqual(execute_payload, {"input": {"orderId": "123"}})

    def test_canonical_error_mapping(self) -> None:
        fake = FakeSession(
            [
                FakeResponse(
                    403,
                    {
                        "ok": False,
                        "data": None,
                        "meta": {
                            "requestId": "req_denied",
                            "timingMs": {"total": 5.1},
                        },
                        "error": {
                            "code": "POLICY_DENIED",
                            "message": "policy denied",
                            "status": 403,
                            "retryable": False,
                        },
                    },
                )
            ]
        )
        client = DriftGateClient(base_url="https://api.driftgate.ai", bearer_token="token", session=fake)

        with self.assertRaises(DriftGateSDKError) as ctx:
            client.session.start(agent="refund-agent")

        self.assertEqual(ctx.exception.code, "POLICY_DENIED")
        self.assertEqual(ctx.exception.status_code, 403)
        self.assertEqual(ctx.exception.correlation_id, "req_denied")


if __name__ == "__main__":
    unittest.main()
